import itertools
import re
from abc import abstractmethod
from collections import Sequence
from string import Template
from typing import TextIO, List, overload, Dict, Tuple
from .helpers import *
from .fuzzyfind import *
from statistics import mean
import operator
import tempfile
import os
import hashlib
from io import StringIO
from collections import Counter
from math import ceil

page_number_template = Template(
    r"(?<!(rule|form|year|ears|tion) )(?<!\d.)(?<![-/,$$:;])(?<!&#)(?<=\D)" + PAGE_NUMBER_TEMPLATE_STR + r"(?=\D)(?![\-:,%/\d])(?!\.(\d|\S))")
tag_attribute_PATTERN = r'(v?align|src|alt|colspan|rowspan|style|cellpadding|id|width|height|(bg)?color|cellspacing|border|face|name|size)=(\'|").*?(\3)'
tag_attribute_RE = re.compile(tag_attribute_PATTERN, flags=re.DOTALL | re.IGNORECASE)

tag_noshade_PATTERN = r'\s(noshade|nowrap)\s?(?=>)?'
tag_noshade_RE = re.compile(tag_noshade_PATTERN, flags=re.DOTALL | re.IGNORECASE)

tag_condense_PATTERN = r'(<\w+)\s*?(>)'
tag_condense_RE = re.compile(tag_condense_PATTERN, flags=re.DOTALL | re.IGNORECASE)

word_PATTEN = r'(?<![\/=])\b[a-zA-Z]{3,}\b(?![\/=])'
word_RE = re.compile(word_PATTEN, re.IGNORECASE)


def clean_document_attributes(dirty_text: List[str]):
    document = list(itertools.dropwhile(lambda x: not re.match('<DOCUMENT>', x, flags=re.IGNORECASE),
                                        dirty_text))
    if len(document) < 1:
        raise Exception("No '<DOCUMENT>' section found.")
    input_end = ''.join(itertools.takewhile(lambda x: not re.match('</DOCUMENT>', x, flags=re.IGNORECASE),
                                            document)) + '</DOCUMENT>'
    input_no_attributes = re.sub(tag_attribute_RE, ' ', input_end)
    input_noshade = re.sub(tag_noshade_RE, ' ', input_no_attributes)
    return StringIO(re.sub(tag_condense_RE, r'\g<1>\g<2>', input_noshade)).readlines()


def discover_new_page_template(page_markers_forward, page_markers_reverse, force_short_discovery=False):
    all_page_numbers = sort_unique([k for k in page_markers_forward.keys()])

    # TODO: Do something real with this aside from throw an exception
    if not force_short_discovery and (len(page_markers_forward) < 20 or len(page_markers_reverse) < 20):
        raise Exception("This shouldn't be an exception, but here we are.")

    if force_short_discovery:
        fifteen_percent_forward = 0
        fifteen_percent_reverse = 0
    else:
        fifteen_percent_forward = ceil((len(page_markers_forward) * .15))
        fifteen_percent_reverse = ceil((len(page_markers_reverse) * .15))

    forward_min = min(all_page_numbers)
    reverse_max = max(all_page_numbers)

    normalized_forward = normalize_page_numbers_to_template(page_markers_forward)
    normalized_reverse = normalize_page_numbers_to_template(page_markers_reverse)

    worsted_forward = {k: v for k, v in normalized_forward.items() if k >= forward_min + fifteen_percent_forward}
    worsted_reverse = {k: v for k, v in normalized_reverse.items() if k <= reverse_max - fifteen_percent_reverse}

    best_match_forward = recurse_through_page_markers(worsted_forward)
    best_match_reverse = recurse_through_page_markers(worsted_reverse)
    if best_match_forward == best_match_reverse:
        return best_match_forward
    else:
        combined_best_match = longest_match_from_list([best_match_forward, best_match_reverse])
        if len(combined_best_match) > len(PAGE_NUMBER_TEMPLATE_STR):
            return combined_best_match
        else:
            return PAGE_NUMBER_TEMPLATE_STR


def recurse_through_page_markers(page_markers: Dict[int, Tuple[int, str]]) -> str:
    if len(page_markers) == 1:
        return list(page_markers.values())[0][1]
    if len(page_markers) >= 2:
        first_dict, second_dict = shuffle_split_dictionary_in_half(page_markers)
        return longest_match_from_list([recurse_through_page_markers(first_dict), recurse_through_page_markers(second_dict)])


def iterative_page_number_search(template, document_slice, known_pages, start_page):
    last_page = start_page - 1
    last_index = 0
    found_page = True
    while found_page:
        page_number = last_page + 1
        page_find_regex = template.substitute({f"{PAGE_NUMBER_NAME}": page_number})

        for line_number, line in enumerate(document_slice[last_index:], last_index):
            found_page = re.search(page_find_regex, line)
            if found_page:
                if line[-1:] == "\n":
                    known_pages[page_number] = (line_number, line[:-1])
                elif line[-2:] == "\r\n":
                    known_pages[page_number] = (line_number, line[:-2])
                else:
                    known_pages[page_number] = (line_number, line)

                last_index = line_number
                last_page = page_number
                break

    return known_pages


def reversed_sliced_page_number_search(template, forward_document_slice, known_pages, start_page, end_page, last_index):
    last_page = end_page
    found_page = True
    forward_range = range(last_index, last_index + len(forward_document_slice))
    document_slice = zip(forward_range, forward_document_slice)

    reversed_document_slice = reversed(list(document_slice))
    while found_page and start_page <= last_page:

        page_number = last_page
        page_find_regex = template.substitute({f"{PAGE_NUMBER_NAME}": page_number})
        if template != page_number_template:
            page_find_regex = re.escape(page_find_regex)
        for line_number, line in reversed_document_slice:
            found_page = re.search(page_find_regex, line, re.IGNORECASE)
            if found_page:
                if line[-1:] == "\n":
                    known_pages[page_number] = (line_number, line[:-1])
                elif line[-2:] == "\r\n":
                    known_pages[page_number] = (line_number, line[:-2])
                else:
                    known_pages[page_number] = (line_number, line)
                last_page = page_number - 1
                break

    return known_pages


def retry_missing_pages(search_template, document_slice, missing_page_numbers, known_pages, offset):
    sorted_page_numbers = sort_unique(missing_page_numbers)
    return reversed_sliced_page_number_search(search_template, document_slice, known_pages, sorted_page_numbers[0], sorted_page_numbers[-1], offset)


def match_or_close(page_number, line_marker,  page_markers_forward):
    return (page_number in page_markers_forward and page_markers_forward[page_number][0] == line_marker[0]) or abs(page_markers_forward[page_number][0] - line_marker[0]) < 10


def is_intraline_document(page_markers_forward, page_markers_reverse):
    forward_word_scores = mean([len(word_RE.findall(val[1])) for page, val in page_markers_forward.items()])
    reverse_word_scores = mean([len(word_RE.findall(val[1])) for page, val in page_markers_reverse.items()])

    return forward_word_scores > 4 or reverse_word_scores > 4


def find_forward_and_reverse_page_numbers(page_markers_forward, start_page, template, unmarked_document):
    # Going numerically forward get all page markers
    page_markers_forward = iterative_page_number_search(template, unmarked_document, page_markers_forward, start_page)
    all_page_numbers = sort_unique([k for k in page_markers_forward.keys()])
    # Going backward get all page markers
    page_markers_reverse = reversed_sliced_page_number_search(template, unmarked_document, {}, all_page_numbers[0],
                                                              all_page_numbers[-1], 0)
    return all_page_numbers, page_markers_forward, page_markers_reverse


def discover_pages(unmarked_document, template, start_page, force_short_discovery=False):
    page_markers_forward = {}
    best_match = None
    all_page_numbers, page_markers_forward, page_markers_reverse = find_forward_and_reverse_page_numbers(
        page_markers_forward, start_page, template, unmarked_document)

    if not is_intraline_document(page_markers_forward, page_markers_reverse):
        # Select forward and backward page markers that match line no
        combined_page_markers = {k: v for k, v in page_markers_reverse.items() if match_or_close(k, v, page_markers_forward)}
    else:
        best_match = discover_new_page_template(page_markers_forward, page_markers_reverse, force_short_discovery)
        if best_match == PAGE_NUMBER_TEMPLATE_STR:
            combined_page_markers = {k: v for k, v in page_markers_reverse.items() if
                                     match_or_close(k, v, page_markers_forward)}
        else:
            all_page_numbers, page_markers_forward, page_markers_reverse = find_forward_and_reverse_page_numbers(
                page_markers_forward, start_page, Template(best_match), unmarked_document)
            combined_page_markers = {k: v for k, v in page_markers_reverse.items() if
                                     match_or_close(k, v, page_markers_forward)}

    # Create counts of each type of page marker
    likely_page_markers = Counter()

    if not best_match:
        for i in combined_page_markers.keys():
            discovered_template = combined_page_markers[i][1].replace(str(i), PAGE_NUMBER_TEMPLATE_STR)
            likely_page_markers[discovered_template] = likely_page_markers[discovered_template] + 1
        best_match = max(likely_page_markers.items(), key=operator.itemgetter(1))[0]

        unfound_pages = sorted([k for k in all_page_numbers if k not in combined_page_markers or combined_page_markers[k][1].replace(str(k), PAGE_NUMBER_TEMPLATE_STR) != best_match])
        found_pages = {k: v for k, v in combined_page_markers.items() if k not in unfound_pages}

    else:
        unfound_pages = sorted([k for k in all_page_numbers if k not in combined_page_markers or best_match not in combined_page_markers[k][1].replace(str(k), PAGE_NUMBER_TEMPLATE_STR)])
        found_pages = {k: v for k, v in combined_page_markers.items() if k not in unfound_pages}
    # page_distances = [found_pages[page + 1][0] - found_pages[page][0] for page in all_page_numbers if
    #                   page in found_pages and page + 1 in found_pages]

    # average_lines_per_page = mean(page_distances)

    # Create list of lists for continuous pages
    continuous_sections = section_continuous_numbers(unfound_pages)
    for section in continuous_sections:
        window_start, window_end = get_window_from_found_pages(section, found_pages, len(unmarked_document))

        # Using our best match template, try to find pages in expected window
        found_pages = retry_missing_pages(Template(best_match), unmarked_document[window_start:window_end], section,
                                          found_pages, offset=0)

    # Update list of unfound pages
    unfound_pages = [k for k in unfound_pages if k not in found_pages.keys()]

    continuous_sections = section_continuous_numbers(unfound_pages)
    for section in continuous_sections:
        window_start, window_end = get_window_from_found_pages(section, found_pages, len(unmarked_document))
        document_slice = unmarked_document[window_start:window_end]

        # Use fuzzy matching for any remaining pages that are not found
        for n in section:
            try:
                found_line = fuzzy_find_by_template(document_slice, Template(best_match), n, offset=window_start)
                found_pages[n] = found_line
            except Exception as err:
                continue

    return found_pages


class PaginatedDocument(Sequence):

    def __getitem__(self, s):
        if isinstance(s, slice):
            start, stop = get_window_from_found_pages([s.start, s.stop], self.page_endings, len(self.cleaned))
            return self.cleaned[start:stop]
        elif isinstance(s, int):
            start, stop = get_window_from_found_pages([s, s], self.page_endings, len(self.cleaned))
            return self.cleaned[start:stop]

    def __len__(self):
        return len(self.page_endings)

    def __init__(self, document: TextIO, start_page: int = 1, clean_xml=False, write_tempfile=False):
        super().__init__()
        self.original = document.readlines()
        self.cleaned = self.original
        if clean_xml:
            self.cleaned = clean_document_attributes(self.original)

        if write_tempfile:
            tempdir = tempfile.gettempdir()
            self.temp_filename = tempdir + os.sep + str(hashlib.sha256(''.join(self.original).encode()).hexdigest())
            with open(self.temp_filename, 'w') as temp:
                temp.writelines(self.cleaned)
        self.page_endings = discover_pages(self.cleaned, page_number_template, start_page)
        self.page_template = self.get_page_template()

    def get_page_template(self):
        likely_page_markers = Counter()
        for i in self.page_endings.keys():
            discovered_template = self.page_endings[i][1].replace(str(i), PAGE_NUMBER_TEMPLATE_STR)
            likely_page_markers[discovered_template] = likely_page_markers[discovered_template] + 1
            return max(likely_page_markers.items(), key=operator.itemgetter(1))[0]


__all__ = ['PaginatedDocument']






from discoverpagination.paginateddocument import discover_pages, is_intraline_document, page_number_template
from string import Template
from unittest import TestCase


class TestDiscoverPages(TestCase):

    def test_discover_pages_find_intraline_number(self):
        # arrange
        document = ["""engendered, probably, by the decomposition of rank vegetable substances\n""",
                    """in a hot and humid soil.[7] The season of the bilious fever,--_vómito_,\n""",
                    """as it is called,--which scourges these coasts, continues from the spring\n""",
                    """to the autumnal equinox, when it is checked by the cold winds that\n"""]


        # action
        result = discover_pages(document, page_number_template, 7, force_short_discovery=True)

        # assert
        self.assertTrue(7 in result)
        pass

    def test_discover_pages_find_intraline_number_as_pattern(self):
        # arrange
        document = ["""engendered, probably, by the decomposition of rank vegetable substances\n""",
                    """in a hot and humid soil.[7] The season of the bilious fever,--_vómito_,\n""",
                    """as it is called,--which scourges these coasts, continues from the spring\n""",
                    """to the autumnal equinox, when it is checked by the cold winds that\n""",
                    """engendered, probably, by the decomposition of rank vegetable substances\n""",
                    """in a hot and humid soil. The season of the bilious fever,--_vómito_,\n""",
                    """as it is called,--which [8]scourges these coasts, continues from the spring\n""",
                    """to the autumnal equinox, when it is checked by the cold winds that\n"""
                    ]

        # action
        result = discover_pages(document, page_number_template, 7, force_short_discovery=True)

        # assert
        self.assertTrue(7 in result)


class TestIsIntralineDocument(TestCase):

    def test_is_intraline_document_true(self):
        # arrange
        page_numbers_forward = {7: (1, 'in a hot and humid soil.[7] The season of the bilious fever,--_vómito_,'),
                                8: (6, 'as it is called,--which [8]scourges these coasts, continues from the spring')}
        page_number_reverse = {8: (6, 'as it is called,--which [8]scourges these coasts, continues from the spring'),
                               7: (1, 'in a hot and humid soil.[7] The season of the bilious fever,--_vómito_,')}

        # action
        result = is_intraline_document(page_numbers_forward, page_number_reverse)

        # assert
        self.assertTrue(result)



from fuzzywuzzy import process
from typing import List
import re
from string import Template
PAGE_NUMBER_PATTERN = r"\b\d+\b"
PAGE_NUMBER_RE = re.compile(PAGE_NUMBER_PATTERN, re.IGNORECASE)


def fuzzy_find_by_template(document_slice: List[str], template: Template, page_number: int, offset: int = 0):
    search_sample = template.substitute({'page_number': page_number})
    possible_lines = [l for l in document_slice if str(page_number) in l and PAGE_NUMBER_RE.search(l)]
    if not possible_lines:
        raise Exception
    ratings = process.extract(search_sample,  possible_lines)
    best_match = max(ratings, key=lambda x: x[1])[0]
    lineno = document_slice.index(best_match)

    if best_match[-1:] == "\n":
        return lineno + offset, best_match[:-1]
    elif best_match[-2:] == "\r\n":
        return lineno + offset, best_match[:-2]
    else:
        return lineno + offset, best_match


__all__ = ['fuzzy_find_by_template']



from math import ceil
from typing import List, Dict, Tuple, TypeVar
from difflib import SequenceMatcher
from random import shuffle

KT = TypeVar('KT')
VT = TypeVar('VT')

PAGE_NUMBER_NAME = r'page_number_of_a_really_long_length'
PAGE_NUMBER_TEMPLATE_STR = r'${' + PAGE_NUMBER_NAME + r'}'


def section_continuous_numbers(arr: List[int]) -> List[List[int]]:
    number_lists = []
    if not arr:
        return []
    sorted_unique = sort_unique(arr)

    current_number = sorted_unique[0] - 1
    current_index = 0
    number_lists.append([])
    for number in sorted_unique:
        if number == current_number + 1:
            number_lists[current_index].append(number)
            current_number = number
        else:
            current_index = current_index + 1
            number_lists.append([])
            number_lists[current_index].append(number)
            current_number = number

    return number_lists


def shuffle_split_dictionary_in_half(d: Dict[KT, VT]) -> Tuple[Dict[KT, VT], Dict[KT, VT]]:
    shuffled_keys = list(d.keys())
    half_length = ceil(len(shuffled_keys) / 2)
    shuffle(shuffled_keys)

    first_half = {k: v for k, v in d.items() if shuffled_keys.index(k) <= half_length - 1}
    second_half = {k: v for k, v in d.items() if shuffled_keys.index(k) > half_length - 1}

    return first_half, second_half


def normalize_page_numbers_to_template(page_number_dict: Dict[int, Tuple[int, str]]) -> Dict[int, Tuple[int, str]]:
    return {k: (v[0], v[1].replace(str(k), PAGE_NUMBER_TEMPLATE_STR)) for k, v in page_number_dict.items()}


def get_window_from_found_pages(section_pages: List[int], found_pages: Dict[int, Tuple[int, str]],
                                document_size: int) -> Tuple[int, int]:
    if not section_pages:
        raise IndexError("section_pages must contain at least 1 value.")

    if not found_pages:
        raise Exception("found_numbers must contain at least 1 value.")

    sorted_unique = sort_unique(section_pages)
    start = sorted_unique[0]
    end = sorted_unique[-1]

    max_found = max([page for page in found_pages.keys()])

    start_line = 0
    if start > 1:
        earliest_Known_pages = {page: t for page, t in found_pages.items() if page < start}.keys()

        if earliest_Known_pages:
            start_page = max(earliest_Known_pages)
            start_line = found_pages[start_page][0]

    end_line = document_size - 1
    if max(section_pages) < max_found:
        end_page = min({page: t for page, t in found_pages.items() if page >= end}.keys())
        end_line = found_pages[end_page][0]

    return start_line, end_line


def longest_match_in_page_marker_pair(page_markers: Dict[int, Tuple[int, str]]) -> str:
    if len(page_markers) != 2:
        raise ValueError("'page_markers' must be a Dict of len 2.")
    normalized_page_numbers = normalize_page_numbers_to_template(page_markers)

    normalized_lines = [v[1] for k, v in normalized_page_numbers.items()]

    return longest_match_from_list(normalized_lines)


def longest_match_from_list(normalized_lines: List[str]) -> str:
    if len(normalized_lines) != 2:
        raise ValueError("'normalized_lines' must be a list of len 2.")

    seq_match = SequenceMatcher(isjunk=None, a=normalized_lines[0], b=normalized_lines[1], autojunk=False)
    match_obj = seq_match.find_longest_match(alo=0, ahi=len(normalized_lines[0]), blo=0, bhi=len(normalized_lines[1]))

    if match_obj.size < 1:
        raise ValueError("No matching strings found, confirm there is a common string template.")

    if match_obj.size < len(PAGE_NUMBER_TEMPLATE_STR):
        raise ValueError("Match length is less than PAGE_NUMBER_TEMPLATE_STR, something is wrong.")

    return normalized_lines[0][match_obj.a:match_obj.a + match_obj.size]


def sort_unique(arr: List[int]) -> List[int]:
    return sorted(list(set(arr)))


def get_lines_range_from_page_range(start_page: int, end_page: int, found_pages: Dict[int, Tuple[int, str]],
                                    document) -> Tuple[int, int]:
    start_line = 0
    end_line = len(document) - 1

    if start_page > 1 and start_page - 1 in found_pages.keys():
        start_line = found_pages[start_page - 1][0]

    if end_page <= max([x for x, y in found_pages.items()]) and end_page in found_pages.keys():
        end_line = found_pages[end_page][0]

    if start_page - 1 not in found_pages.keys() or end_page not in found_pages.keys():
        return get_window_from_found_pages([start_page, end_page], found_pages, document_size=len(document))

    return start_line, end_line


__all__ = ['section_continuous_numbers', 'get_window_from_found_pages', 'sort_unique', 'get_lines_range_from_page_range', 'PAGE_NUMBER_TEMPLATE_STR', 'normalize_page_numbers_to_template', 'longest_match_in_page_marker_pair', 'longest_match_from_list',
           'shuffle_split_dictionary_in_half', 'PAGE_NUMBER_NAME']



import setuptools

setuptools.setup(
    name='DiscoverPagination',
    version='0.1.4',
    packages=setuptools.find_packages(),
    license='MIT',
    author_email="dhking@wharton.upenn.edu",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=['fuzzywuzzy', 'python-Levenshtein'],
)



from unittest import TestCase
from discoverpagination.helpers import *


class TestSplitDictionaryInHalf(TestCase):

    def test_split_dictionary_in_half_len_2(self):
        # arrange
        subject = {1: "one", 2: "two"}

        # action
        first, second = shuffle_split_dictionary_in_half(subject)

        # assert

        self.assertEqual(2, len(subject))
        self.assertEqual(1, len(first))
        self.assertEqual(1, len(second))

        if 1 not in first:
            self.assertTrue(1 in second)

        if 2 not in first:
            self.assertTrue(2 in second)

        if 1 not in second:
            self.assertTrue(1 in first)

        if 2 not in second:
            self.assertTrue(2 in first)

        if 1 in first:
            self.assertTrue(1 not in second)

        if 2 in first:
            self.assertTrue(2 not in second)

        # No overlapping keys
        self.assertFalse([k for k in first.keys() if k in second.keys()])

    def test_split_dictionary_in_half_len_5(self):
        # arrange
        subject = {1: "one", 2: "two", 3: "three", 4: "four", 5: "five"}

        # action
        first, second = shuffle_split_dictionary_in_half(subject)

        # assert
        self.assertEqual(5, len(subject))
        self.assertEqual(3, len(first))
        self.assertEqual(2, len(second))

        # No overlapping keys
        self.assertFalse([k for k in first.keys() if k in second.keys()])

    def test_split_dictionary_in_half_len_6(self):
        # arrange
        subject = {1: "one", 2: "two", 3: "three", 4: "four", 5: "five", 6: "six"}

        # action
        first, second = shuffle_split_dictionary_in_half(subject)

        # assert
        self.assertEqual(6, len(subject))
        self.assertEqual(3, len(first))
        self.assertEqual(3, len(second))

        # No overlapping keys
        self.assertFalse([k for k in first.keys() if k in second.keys()])

    def test_split_dictionary_in_half_len_1(self):
        # arrange
        subject = {1: "one"}

        # action
        first, second = shuffle_split_dictionary_in_half(subject)

        # assert
        self.assertEqual(1, len(subject))
        self.assertEqual(1, len(first))
        self.assertEqual(0, len(second))

        # No overlapping keys
        self.assertFalse([k for k in first.keys() if k in second.keys()])

    def test_split_dictionary_in_half_len_0(self):
        # arrange
        subject = {}

        # action
        first, second = shuffle_split_dictionary_in_half(subject)

        # assert
        self.assertEqual(0, len(subject))
        self.assertEqual(0, len(first))
        self.assertEqual(0, len(second))

        # No overlapping keys
        self.assertFalse([k for k in first.keys() if k in second.keys()])


class TestSectionContinuousNumbers(TestCase):

    def test_section_continuous_numbers_make_continuous_sublists(self):
        # arrange
        arr = [1, 1, 1, 1, 1, 2, 3, 5, 5, 5]

        # action
        result = section_continuous_numbers(arr)

        # assert
        self.assertTrue(isinstance(result, list))

        self.assertEqual(2, len(result))

        self.assertEqual(1, result[0][0])
        self.assertEqual(5, result[1][0])

    def test_section_continuous_numbers_make_three_lists_of_one_number(self):
        # arrange
        arr = [1, 3, 5]

        # action
        result = section_continuous_numbers(arr)

        # assert
        self.assertTrue(isinstance(result, list))
        self.assertEqual(3, len(result))
        self.assertEqual(1, result[0][0])
        self.assertEqual(3, result[1][0])
        self.assertEqual(5, result[2][0])

    def test_section_continuous_numbers_make_three_lists_of_three_numbers(self):
        # arrange
        arr = [1, 2, 3, 5, 6, 7, 21, 20, 19]

        # action
        result = section_continuous_numbers(arr)

        # assert
        self.assertTrue(isinstance(result, list))

        self.assertEqual(3, len(result))
        self.assertEqual(3, len(result[0]))
        self.assertEqual(3, len(result[1]))
        self.assertEqual(3, len(result[2]))

        self.assertEqual(1, result[0][0])
        self.assertEqual(3, result[0][2])

        self.assertEqual(5, result[1][0])
        self.assertEqual(7, result[1][2])

        self.assertEqual(19, result[2][0])
        self.assertEqual(21, result[2][2])


class TestGetWindowFromFoundPages(TestCase):
    #
    # def test_get_window_from_found_pages_reversed_start_end_error(self):
    #     # arrange
    #     arr = [16]
    #     found_pages = {6: (984, '    6'), 7: (2904, '    7'), 8: (3036, '    8'), 9: (3212, '    9'),
    #                    10: (3447, '    10'),
    #                    11: (3582, '    11'), 12: (4346, '    12'), 13: (5271, '    13'), 14: (5485, '    14'),
    #                    15: (5526, '    15'),
    #                    20: (8958, '    20'), 21: (9105, '    21'), 22: (9237, '    22'), 23: (9452, '    23'),
    #                    24: (9731, '    24'),
    #                    25: (9969, '    25'), 26: (10403, '    26'), 27: (10531, '    27'), 28: (10662, '    28'),
    #                    29: (10801, '    29'),
    #                    30: (10895, '    30'), 31: (13467, '    31'), 32: (14201, '    32'), 33: (14372, '    33'),
    #                    34: (14448, '    34'),
    #                    35: (17830, '    35'), 36: (18010, '    36'), 37: (18036, '    37'), 38: (22863, '    38'),
    #                    39: (24824, '    39'),
    #                    40: (26515, '    40'), 41: (27785, '    41'), 42: (27958, '    42'), 43: (28166, '    43'),
    #                    44: (29046, '    44'),
    #                    45: (32372, '    45'), 46: (32545, '    46'), 47: (32725, '    47'), 48: (32884, '    48'),
    #                    49: (33034, '    49'),
    #                    50: (33175, '    50'), 51: (33319, '    51'), 52: (33573, '    52'), 53: (33713, '    53'),
    #                    54: (34060, '    54'),
    #                    55: (34202, '    55'), 56: (34336, '    56'), 57: (34391, '    57'), 58: (34463, '    58'),
    #                    59: (34668, '    59'),
    #                    60: (34806, '    60'), 61: (34855, '    61'), 62: (35043, '    62'), 63: (35138, '    63'),
    #                    64: (35268, '    64'),
    #                    19: (1795, '    19'), 18: (1531, '    18'), 17: (1286, '    17')}
    #
    #     # action
    #     start, end = get_window_from_found_pages(arr, found_pages, document_size=110)
    #
    #     # assert
    #     self.assertEqual(0, start)

    def test_get_window_from_found_pages_first_page_equals_zero(self):
        # arrange
        arr = [1]
        found_pages = {2: (20, "ok"), 3: (30, "ok"), 4: (40, "ok"), 5: (50, "ok"), 6: (60, "ok"), 7: (70, "ok"),
                       8: (80, "ok"), 11: (110, "ok")}

        # action
        start, end = get_window_from_found_pages(arr, found_pages, document_size=110)

        # assert
        self.assertEqual(0, start)

    def test_get_window_from_found_pages_first_page_equals_zero_gt_1(self):
        # arrange
        arr = [2]
        found_pages = {3: (30, "ok"), 4: (40, "ok"), 5: (50, "ok"), 6: (60, "ok"), 7: (70, "ok"),
                       8: (80, "ok"), 11: (110, "ok")}

        # action
        start, end = get_window_from_found_pages(arr, found_pages, document_size=110)

        # assert
        self.assertEqual(0, start)
    def test_get_window_from_found_pages_end_page_equals_document_size(self):
        # arrange
        arr = [9]
        found_pages = {1: (10, "ok"), 2: (20, "ok"), 3: (30, "ok"), 4: (40, "ok"), 5: (50, "ok"), 6: (60, "ok"),
                       7: (70, "ok"), 8: (80, "ok")}

        # action
        start, end = get_window_from_found_pages(arr, found_pages, document_size=234)

        # assert
        self.assertEqual(233, end)

    def test_get_window_from_found_pages_end_page_beyond_found(self):
        # arrange
        arr = [123]
        found_pages = {1: (10, "ok"), 2: (20, "ok"), 3: (30, "ok"), 4: (40, "ok"), 5: (50, "ok"), 6: (60, "ok"),
                       7: (70, "ok"), 8: (80, "ok")}

        # action
        start, end = get_window_from_found_pages(arr, found_pages, document_size=234)

        # assert
        self.assertEqual(233, end)

    def test_get_window_from_found_pages_start_end_index_single_page(self):
        # arrange
        arr = [9]
        found_pages = {1: (10, "ok"), 2: (20, "ok"), 3: (30, "ok"), 4: (40, "ok"), 5: (50, "ok"), 6: (60, "ok"),
                       7: (70, "ok"), 8: (80, "ok"), 11: (110, "ok")}

        # action
        start, end = get_window_from_found_pages(arr, found_pages, document_size=120)

        # assert

        self.assertEqual(80, start)
        self.assertEqual(110, end)

    def test_get_window_from_found_pages_start_end_index_multi_page(self):
        # arrange
        arr = [4, 5, 6, 7, 8, 9, 10]
        found_pages = {1: (10, "ok"), 2: (20, "ok"), 3: (30, "ok"), 11: (111, "ok"), 12: (120, "ok"), 13: (130, "ok"), 14: (140, "ok")}

        # action
        start, end = get_window_from_found_pages(arr, found_pages, document_size=150)

        # assert

        self.assertEqual(30, start)
        self.assertEqual(111, end)


class TestSortUnique(TestCase):
    def test_sort_unique_empty_array_returns_empty(self):
        # arrange
        arr = []

        # action
        result = sort_unique(arr)

        # assert
        self.assertTrue(isinstance(result, list))
        self.assertEqual(0, len(arr))

    def test_sort_unique_sort_simple(self):
        # arrange
        arr = [1, 2, 3]

        # action
        result = sort_unique(arr)

        # assert
        self.assertTrue(isinstance(result, list))

        self.assertEqual(3, len(result))

        self.assertEqual(1, result[0])
        self.assertEqual(2, result[1])
        self.assertEqual(3, result[2])

    def test_sort_unique_sort_complex(self):
        # arrange
        arr = [3, 1, 2, 10, 3, 6, 8, 8, 9, 10]

        # action
        result = sort_unique(arr)

        # assert
        self.assertTrue(isinstance(result, list))

        self.assertEqual(7, len(result))

        self.assertEqual(1, result[0])
        self.assertEqual(2, result[1])
        self.assertEqual(10, result[6])

    def test_sort_unique_make_unique(self):
        # arrange
        arr = [1, 1, 1, 1, 1, 5, 5, 5, 5, 5]

        # action
        result = sort_unique(arr)

        # assert
        self.assertTrue(isinstance(result, list))

        self.assertEqual(2, len(result))

        self.assertEqual(1, result[0])
        self.assertEqual(5, result[1])


class TestGetLinesRangeFromPageRange(TestCase):
    def test_get_lines_range_from_page_range_two_pages(self):
        # arrange
        found_pages = {1: (10, "ok"), 2: (20, "ok"), 3: (30, "ok")}
        start_page = 2
        end_page = 3
        document = ['Lorem ipsum'] * 31
        # action
        start_line, end_line = get_lines_range_from_page_range(start_page=start_page, end_page=end_page, found_pages=found_pages, document=document)

        # assert
        self.assertEqual(10, start_line)
        self.assertEqual(30, end_line)

    def test_get_lines_range_from_page_range_ten_pages(self):
        # arrange
        found_pages = {1: (10, "ok"), 2: (20, "ok"), 3: (30, "ok"), 4: (40, "ok"), 5: (50, "ok"), 6: (60, "ok"), 7: (70, "ok"), 8: (80, "ok"),
                       9: (90, "ok"), 10: (100, "ok"), 11: (110, "ok"), 12: (120, "ok"), 13: (130, "ok"), 14: (140, "ok"), 15: (150, "ok"), 16: (160, "ok")}

        start_page = 4
        end_page = 14
        document = ['Lorem ipsum'] * 161
        # action
        start_line, end_line = get_lines_range_from_page_range(start_page=start_page, end_page=end_page, found_pages=found_pages, document=document)

        # assert
        self.assertEqual(30, start_line)
        self.assertEqual(140, end_line)

    def test_get_lines_range_from_page_range_where_start_is_missing(self):
        # arrange
        found_pages = {1: (10, "ok"), 2: (20, "ok"), 9: (90, "ok"), 10: (100, "ok"), 11: (110, "ok"), 12: (120, "ok"), 13: (130, "ok"), 14: (140, "ok"), 15: (150, "ok"), 16: (160, "ok")}

        start_page = 4
        end_page = 14
        document = ['Lorem ipsum'] * 161
        # action
        start_line, end_line = get_lines_range_from_page_range(start_page=start_page, end_page=end_page, found_pages=found_pages, document=document)

        # assert
        self.assertEqual(20, start_line)
        self.assertEqual(140, end_line)

    def test_get_lines_range_from_page_range_where_end_is_missing(self):
        # arrange
        found_pages = {1: (10, "ok"), 2: (20, "ok"), 3: (30, "ok"), 4: (40, "ok"), 5: (50, "ok"), 6: (60, "ok"), 7: (70, "ok"), 8: (80, "ok"),
                       16: (160, "ok")}

        start_page = 4
        end_page = 14
        document = ['Lorem ipsum'] * 161
        # action
        start_line, end_line = get_lines_range_from_page_range(start_page=start_page, end_page=end_page,
                                                               found_pages=found_pages, document=document)

        # assert
        self.assertEqual(30, start_line)
        self.assertEqual(160, end_line)


class TestLongestMatchFromList(TestCase):

    def test_longest_match_from_list_simple(self):
        # arrange
        test_list = [PAGE_NUMBER_TEMPLATE_STR, PAGE_NUMBER_TEMPLATE_STR]

        # action
        result = longest_match_from_list(test_list)

        # assert
        expected = PAGE_NUMBER_TEMPLATE_STR
        self.assertEqual(expected, result)

    def test_longest_match_from_list_complex(self):
        # arrange
        test_list = ['They established their capital at Tula ' + PAGE_NUMBER_TEMPLATE_STR + ', north of the Mexican', '' + PAGE_NUMBER_TEMPLATE_STR + " appearance of almost modern constructions."]

        # action
        result = longest_match_from_list(test_list)

        # assert
        expected = PAGE_NUMBER_TEMPLATE_STR
        self.assertEqual(expected, result)

    def test_longest_match_from_list_extremities(self):
        # arrange
        test_list = ['They established their capital at Tula ' + PAGE_NUMBER_TEMPLATE_STR, PAGE_NUMBER_TEMPLATE_STR + " appearance of almost modern constructions."]

        # action
        result = longest_match_from_list(test_list)

        # assert
        expected = PAGE_NUMBER_TEMPLATE_STR
        self.assertEqual(expected, result)

    def test_longest_match_from_list_long(self):
        # arrange
        test_list = ["---ABCDE" + PAGE_NUMBER_TEMPLATE_STR + "FGHI---", "---ABCDE" + PAGE_NUMBER_TEMPLATE_STR + "FGHI---"]

        # action
        result = longest_match_from_list(test_list)

        # assert
        expected = "---ABCDE" + PAGE_NUMBER_TEMPLATE_STR + "FGHI---"
        self.assertEqual(expected, result)

    def test_longest_match_from_list_complex_long(self):
        # arrange
        test_list = ["They established their capital at Tula ---ABCDE" + PAGE_NUMBER_TEMPLATE_STR + "FGHI---, north of the Mexican", "appearance of almost modern---ABCDE" + PAGE_NUMBER_TEMPLATE_STR + "FGHI--- constructions."]

        # action
        result = longest_match_from_list(test_list)

        # assert
        expected = "---ABCDE" + PAGE_NUMBER_TEMPLATE_STR + "FGHI---"
        self.assertEqual(expected, result)


class TestLongestMatchInPageMarkerPair(TestCase):

    def test_longest_match_in_page_marker_pair_simple(self):
        # arrange
        page_markers_forward = {19: (1475, '19'),
                                22: (1484, '22')}

        # action
        result = longest_match_in_page_marker_pair(page_markers_forward)

        # assert
        expected = PAGE_NUMBER_TEMPLATE_STR
        self.assertEqual(expected, result)

    def test_longest_match_in_page_marker_pair_long(self):
        # arrange
        page_markers_forward = {
            19: (1475, 'times.asd19dsa They established their capital at Tula, north of the Mexican'),
            22: (1484, 'appearance of almost modern constructions.fgh22hgf')}

        # action
        result = longest_match_in_page_marker_pair(page_markers_forward)

        # assert
        expected = PAGE_NUMBER_TEMPLATE_STR
        self.assertEqual(expected, result)

    def test_longest_match_in_page_marker_pair_extremities(self):
        # arrange
        page_markers_forward = {
            19: (1475, '19asddsatimes. They established their capital at Tula, north of the Mexican'),
            22: (1484, 'appearance of almost modern constructions.fghhgf22')}

        # action
        result = longest_match_in_page_marker_pair(page_markers_forward)

        # assert
        expected = PAGE_NUMBER_TEMPLATE_STR
        self.assertEqual(expected, result)

    def test_longest_match_in_page_marker_pair_complex(self):
        # arrange
        page_markers_forward = {
            19: (1475, 'times.[19] They established their capital at Tula, north of the Mexican'),
            22: (1484, 'appearance of almost modern constructions.[22]')}

        # action
        result = longest_match_in_page_marker_pair(page_markers_forward)

        # assert
        expected = 's.[' + PAGE_NUMBER_TEMPLATE_STR + ']'
        self.assertEqual(expected, result)

    def test_longest_match_in_page_marker_pair_complex_long(self):
        # arrange
        page_markers_forward = {
            19: (1475, 'times. --ABCD[19]EFGH-- They established their capital at Tula, north of the Mexican'),
            22: (1484, 'appearance of almost modern constructions. --ABCD[22]EFGH--')}

        # action
        result = longest_match_in_page_marker_pair(page_markers_forward)

        # assert
        expected = 's. --ABCD[' + PAGE_NUMBER_TEMPLATE_STR + ']EFGH--'
        self.assertEqual(expected, result)
